/*******************************************************************************
*   Author       -  Kiraly Tibor
*                   http://www.tkiraaly.hu
*   Date         -  2014.08.12.
*   Chip         -  Atmel ATmega8 & HD44780
*   Compiler     -  avr-gcc (WinAVR)
*
*   LCDT01 - LCD terminal
*   ATmega8, 2x16 karakteres LCD, 3x4 gombos tasztatura, LED, csipogo 
*   
*   Ez egy hagyomanyos soros vonali terminal. A soros vonalon erkezo betuket
*   kiirja a kepernyore, a lenyomott billentyuk kodjat pedig elkuldi a soros
*   vonalon. Kepes a magyar kis betuket is kiirni (WIN kodolas szerint).
*   A soros vonal parameterei: 9800 Baud, 8 bit, No parity, 1 stop bit.   
*   Ertelmezett parancsok:
*   ESC, T    - kepernyo torlese
*   ESC, S    - csipogas
*   ESC, L    - LED bekapcsolasa
*   ESC, K    - LED kikapcsolasa
*   ESC, X    - felso sor kijelolese
*   ESC, Y    - also sor kijelolese
*   ESC, 0..F - kurzor mozgatasa a kijelolt sor adott poziciojaba 
*   
********************************************************************************
*   PonyProg Configuration and Security Bits (bepipalva):
*
*   CKSEL3, CKSEL2, CKSEL1, CKSEL0
*   P       P       -       -       0011  Calibrated Internal RC Oscillator 4MHz
*   P       -       P       P       0100  Calibrated Internal RC Oscillator 8MHz
*
*   Calibrated Internal RC Oscillator
*   SUT1  SUT0

*   -     P         Slowly rising power
*   
********************************************************************************
*    LCD:
*    14   LCD D7       - AVR PD7 - 13
*    13   LCD D6       - AVR PD6 - 12
*    12   LCD D5       - AVR PD5 - 11
*    11   LCD D4       - AVR PD4 -  6
*    10   LCD D3
*     9   LCD D2
*     8   LCD D1
*     7   LCD D0
*     6   LCD E        - AVR PD3 - 5
*     5   LCD RW       - GND
*     4   LCD RS       - AVR PD2 - 4
*     3   LCD KONTR    - 10 kOhm trim.
*     2   LCD VDD      - +5V
*     1   LCD VSS      - GND
*     
*    KBD:
*     1 KBD COL1       - AVR PC0 - 23
*     2 KBD COL2       - AVR PC1 - 24
*     3 KBD COL3       - AVR PC2 - 25
*     4 KBD ROW1       - AVR PD7 - 13 - 4K7      
*     5 KBD ROW2       - AVR PD6 - 12 - 4K7      
*     6 KBD ROW3       - AVR PD5 - 11 - 4K7      
*     7 KBD ROW4       - AVR PD4 - 10 - 4K7      
*
*    Soros port:
*     1 VCC
*     2 X
*     3 RX             - AVR RXD  - 2
*     4 TX             - ACR TXD  - 3
*     5 GND
*     
*    Programozo fej:
*     1 RST            - AVR NRES - 1
*     2 X
*     3 SCK            - AVR SCK  - 19
*     4 MISO           - AVR MISO - 18
*     5 MOSI           - AVR MOSI - 17
*     6 GND
*
*    Egyeb:
*     LED              - AVR PB0 - 14
*     Csipogo          - AVR PC3 - 26
*     I2C CLOCK        - AVR SCL - 28
*     I2C DATA         - AVR SDA - 27
*                      - AVR PB1 - 15
*                      - AVR PB2 - 16
*                      - AVR PB6 - 9
*                      - AVR PB7 - 10
*
*******************************************************************************/

#define F_CPU                _4MHZ
#define T0_DIVIDER           1024                // 0,256 msec




#include "tkiraaly_atmega8.h"
#include <util/delay.h>
#include <avr/pgmspace.h>
#include "tkiraaly_lcddef.h"





// LED - PB0
#define LED_ENABLE           PB0_OUT
#define LED_BE               BC( PORTB, 0)
#define LED_KI               BS( PORTB, 0)

// LCD_E - PD3
#define LCD_E_ENABLE         PD3_OUT
#define LCD_E_0              BC( PORTD, 3)
#define LCD_E_1              BS( PORTD, 3)

// LCD_RS - PD2
#define LCD_RS_ENABLE        PD2_OUT
#define LCD_RS_UTASITAS      BC( PORTD, 2)
#define LCD_RS_ADAT          BS( PORTD, 2)

// LCD adat port - PORTD felso 4 bit - KBD sorok is
#define LCD_PORT             PORTD
#define LCD_PORT_ENABLE      DDRD|= 0B11110000
#define LCD_PORT_DISABLE     DDRD&= 0B00001111


// SPEAKER
#define SPK_ENABLE           PC3_OUT
#define SPK_BE               BC( PORTC, 3)
#define SPK_KI               BS( PORTC, 3)




void lcd_init4( void);                           // LCD inicializalasa 4 bitre
void lcd_putc( U8);                              // betu kiirasa
void lcd_putcmd( U8);                            // parancskod kiadasa
void lcd_yx( U8, U8);                            // kurzor pozicioja 0..3/0..15
void lcd_cls( void);                             // kepernyo torles
void lcd_puts( const char *);                    // string kiirasa
U8 kbd( void);                                   // billentyuzet lekerdezese es dekodolasa
U8 kbd_antiprell( U8);                           // billentyuzet perges mentesito
void usart_putc( U8);                            // USART egy karakter kuldese
void lcd_putc_hu( U8);                           // betu kiirasa, magyar Win konverzio
void lcd_defc( const uint8_t *);                 // egyedi betu(k) beallitasa  
void cmd( U8);                                   // bejovo char ertelmezese
void beep( void);                                // 1000 HZ-es csippantas




const uint8_t magyar_betuk[] PROGMEM =
{
   0,
                                                 //  - 0
   Cx______X_,
   Cx_____X__,
   Cx____XXX_,
   Cx_______X,
   Cx____XXXX,
   Cx___X___X,
   Cx____XXXX,
   Cx________,
                                                 //  - 1
   Cx______X_,
   Cx_____X__,
   Cx____XXX_,
   Cx___X___X,
   Cx___XXXXX,
   Cx___X____,
   Cx____XXX_,
                                                 // - 2
   Cx________,
   Cx______X_,
   Cx_____X__,
   Cx________,
   Cx____XX__,
   Cx_____X__,
   Cx_____X__,
   Cx____XXX_,
   Cx________,
                                                 //  - 3
   Cx______X_,
   Cx_____X__,
   Cx____XXX_,
   Cx___X___X,
   Cx___X___X,
   Cx___X___X,
   Cx____XXX_,
   Cx________,
                                                 //  - 4
   Cx_____X_X,
   Cx____X_X_,
   Cx____XXX_,
   Cx___X___X,
   Cx___X___X,
   Cx___X___X,
   Cx____XXX_,
   Cx________,
                                                 //  - 5
   Cx______X_,
   Cx_____X__,
   Cx________,
   Cx___X___X,
   Cx___X___X,
   Cx___X__XX,
   Cx____XX_X,
   Cx________,
                                                 //  - 6
   Cx_____X_X,
   Cx____X_X_,
   Cx________,
   Cx___X___X,
   Cx___X___X,
   Cx___X__XX,
   Cx____XX_X,
   Cx________,
   
   0xFF
};




volatile U8 speaker= 0;                           // csipogas hossza msec-ben



 
ISR( USART_RXC_vect)                             // USART megszakitas kezelese
{
   cmd( UDR);                                    // bejovo char feldolgozasa
}




ISR( TIMER0_OVF_vect)                            // masodpercenkent kb 100*
{
   U8 c;
   T0_T(10000);                                  // 10000usec - 100Hz
   c= kbd_antiprell( kbd());                     // billentyuzet lekerdezese
   if( c)                                        // ha nyomtak egy gombot
   {
      usart_putc( c);                            // char kuldese
      speaker= 30;                               // billentyu hang
   }
}




int main( void)
{
   SPK_ENABLE;
   LED_ENABLE;
   LED_BE;
   USART_SET_9600_8N1;
   T0_F(100);
   T0_NORMAL;
   T0_IT_OVF_ENABLE;
   IT_ENABLE;
   lcd_init4();
   lcd_cls();
   lcd_defc( magyar_betuk);
   lcd_yx( 0, 0);
   lcd_puts( PSTR( "tkiraaly.hu 2014"));
   lcd_yx( 1, 2);
   lcd_puts( PSTR( "LCD Terminal"));
   _delay_ms( 1000);
   lcd_cls();
   for(;;)
   {
      if( speaker)
      {
         SPK_BE;
         _delay_us( 500);
         SPK_KI;
         _delay_us( 500);
         speaker--;
      }
   }
   return 0;
}




void lcd_init4( void)                            // LCD inicializalasa 4 bitre
{
   LCD_PORT_ENABLE;
   LCD_E_ENABLE;
   LCD_RS_ENABLE;
   _delay_ms( 15);
   LCD_RS_UTASITAS;
   LCD_PORT= 0B00100000;                         // 4 bit interface
   LCD_E_1;
   LCD_E_0;
   _delay_ms( 5);
   LCD_E_1;
   LCD_E_0;
   _delay_us( 120);
   lcd_putc( 0B00101000);                        // 4 bit interface, 2 sor, 5x8 pontos betu
   lcd_putc( 0B00101000);                        // 2x kell kiadni, vagy LCD D3-t VDD-re kell kotni
   lcd_putc( 0B00001100);                        // kijelzes be, cursor ki
   lcd_putc( 0B00000110);                        // kiiras jobbra
   LCD_RS_ADAT;
}




void lcd_putcmd( U8 cmd)                         // parancskod kiadasa
{
   LCD_RS_UTASITAS;
   lcd_putc( cmd);
   LCD_RS_ADAT;
}




void lcd_putc( U8 c)                             // egy betu kiiras
{
   LCD_E_1;                                      // felso 4 bit
   LCD_PORT= ( LCD_PORT & 0x0F) | ( 0xF0 & c);
   LCD_E_0;
   LCD_E_1;                                      // also 4 bit
   LCD_PORT= ( LCD_PORT & 0x0F) | ( 0xF0 & c << 4);
   LCD_E_0;
   _delay_us( 37);                               // var 37 usec  
}




void lcd_yx( U8 sor, U8 betu)                    //  kurzor pozicionalasa 0..3/0..15
{
   U8 cim= 0x80;                                 // parancs kodja
   if (sor & 0B00000001) cim+= 64;               // 1. es 3. sor
   if (sor & 0B00000010) cim+= 20;               // 2. es 3. sor
   cim+= betu & 0x0F; 
   lcd_putcmd( cim);
}




void lcd_cls( void)                              // kepernyo torles
{
   lcd_putcmd( 0x01);
   _delay_ms( 2);
}




void lcd_puts( const char *s)                    // string kiiras
{
   register unsigned char c;
   while ( ( c= pgm_read_byte( s++))) lcd_putc( c);
}




U8 kbd( void)                                    // billentyuzet lekerdezese es dekodolasa
{
   U8 c;
   c= 0;
   LCD_PORT_DISABLE;
   LCD_PORT &= 0x0F;                             // LCD es KBD port reszben kozos
   PD7_OUT;                                      // 1. sor
   NOP;                                          // 4MHz-nel 1 NOP-pal mar jo
   if( BTC( PINC, 0)) c= '1';
   if( BTC( PINC, 1)) c= '2';
   if( BTC( PINC, 2)) c= '3';
   LCD_PORT_DISABLE;
   PD6_OUT;                                      // 2. sor
   NOP;
   if( BTC( PINC, 0)) c= '4';
   if( BTC( PINC, 1)) c= '5';
   if( BTC( PINC, 2)) c= '6';
   LCD_PORT_DISABLE;
   PD5_OUT;                                      // 3. sor
   NOP;
   if( BTC( PINC, 0)) c= '7';
   if( BTC( PINC, 1)) c= '8';
   if( BTC( PINC, 2)) c= '9';
   LCD_PORT_DISABLE;
   PD4_OUT;                                      // 4. sor
   NOP;
   if( BTC( PINC, 0)) c= '*';
   if( BTC( PINC, 1)) c= '0';
   if( BTC( PINC, 2)) c= '#';
   LCD_PORT_ENABLE;
   return c;
}




U8 kbd_antiprell( U8 c)                          // billentyuzet perges mentesito
{
   static U8 n= 0;                               // nyomvatartas szamlalasahoz 
   U8 a= 0;                                      // visszadott ertek, default 0
   if( c) n++;                                   // ha van lenyomott gomb, szamlalas
   else n= 0;
   if( n == 10) n= 9;                            // szamlalo max 9
   if( n == 3) a= c;                             // harmadik periodusnal megy a gomb
   return a;
}




void usart_putc( U8 c)                           // USART egy karakter kuldese
{
    while( BTC( UCSRA, UDRE));                   // 1, ha kesz az adat fogadasara
    UDR= c;
}




void lcd_putc_hu( U8 c)                          // magyar betu kiirasa LCD-re
{
   switch( c)
   {
      case 0xD6:                                 // 
      case 0xF6:                                 // 
                  c= 0xEF;
                  goto kiir;
      case 0xDC:                                 // 
      case 0xFC:                                 // 
                  c= 0xF5;
                  goto kiir;
      case 0xC1:                                 // 
      case 0xE1:                                 // 
                  c= 0;
                  goto kiir;
      case 0xC9:                                 // 
      case 0xE9:                                 // 
                  c= 1;
                  goto kiir;
      case 0xCD:                                 // 
      case 0xED:                                 // 
                  c= 2;
                  goto kiir;
      case 0xD3:                                 // 
      case 0xF3:                                 // 
                  c= 3;
                  goto kiir;
      case 0xD5:                                 // 
      case 0xF5:                                 // 
                  c= 4;
                  goto kiir;
      case 0xDA:                                 // 
      case 0xFA:                                 // 
                  c= 5;
                  goto kiir;
      case 0xDB:                                 // 
      case 0xFB:                                 // 
                  c= 6;
   }
   kiir:
   lcd_putc( c);
}




void cmd( U8 c)                                  // bejovo char ertelmezese
{
   static U8 p= 0;                               // elozo bejovo char
   static U8 sor= 0;
   if( p == 27)                                  // ha az elozo ESC volt
   {
      switch( c)
      {
         case '0':
         case '1':
         case '2':
         case '3':
         case '4':
         case '5':
         case '6':
         case '7':
         case '8':
         case '9':
                   c-= 48;
                   lcd_yx( sor, c);
                   goto vege;
         case 'A':
         case 'B':
         case 'C':
         case 'D':
         case 'E':
         case 'F':
                   c-= 55;
                   lcd_yx( sor, c);
                   goto vege;
         case 'a':
         case 'b':
         case 'c':
         case 'd':
         case 'e':
         case 'f':
                   c-= 87;
                   lcd_yx( sor, c);
                   goto vege;
         case 'X':
         case 'x':
                   sor= 0;
                   goto vege;
         case 'Y':
         case 'y':
                   sor= 1;
                   goto vege;
         case 'T':                               // kepernyo torles
         case 't':
                   lcd_cls();
                   goto vege;
         case 'L':
         case 'l':
                   LED_BE;
                   goto vege;
         case 'K':
         case 'k':
                   LED_KI;
                   goto vege;
         case 'S':                               // csipogas
         case 's':
                   speaker= 100;
                   goto vege;
      }
   }
   if( c != 27) lcd_putc_hu( c);
   vege:
   p= c;
}    




void lcd_defc( const uint8_t *s)                 // egyedi betu(k) beallitasa
{
   register unsigned char c;
   LCD_RS_UTASITAS;
   lcd_putc( ( ( pgm_read_byte( s++) & 0B00000111) << 3) | 0B01000000);   // betu cime + parancs 
   LCD_RS_ADAT;
   while ( ( c= pgm_read_byte( s++)) != 0xFF) lcd_putc( c);
   LCD_RS_UTASITAS;
   LCD_RS_ADAT;
}
